//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// 8k asynchronous FIFO
//
// 
`include "parms.v"

module fifo (
  wclk_i, rclk_i, rst_i, we_i, rd_i, data_i,
  data_o, ffull_o, fempty_o,  fifo_count_o, wack_o, rack_o,
  afull_o, aempty_o 
  );
  input wclk_i;                       // Input write clock signal
  input rclk_i;                       // Input read clock signal
  input rst_i;                        // Input reset signal
  input we_i;                         // Write enable input signal
  input rd_i;                         // Read memory signal
  input [31:0] data_i;                // Input data
  output [31:0] data_o;               // Output data
  output ffull_o;                     // FIFO full output signal
  output fempty_o;                    // FIFO empty output signal
  output [15:0] fifo_count_o;         // Number of 32-bit words in FIFO   
  output wack_o;                      // Write acknowledge handshake signal
  output rack_o;                      // Read acknowledge signal
  output afull_o;                     // Almost full
  output aempty_o;                    // Almost empty
  
  reg [31:0] data_o;
  reg wack_o;
  reg rack_o;
  reg afull_o;
  reg aempty_o;
  reg [15:0] fifo_count_o;
  
 // Internal wires and registers 
  reg [31:0] tfifo [0:2047];    // Tx FIFO 
  reg [11:0] adrr;              // Read address, one extra bit for full/empty detection
  reg [11:0] adrr_s;            // Synchronized read pointer
  reg [11:0] adrw;              // Write address, one extra bit for full/empty detection
  reg [11:0] adrw_s;            // Synchronized write pointer
  reg tfifo_full;               // Flag that FIFO is full
  reg tfifo_empty;              // Flag that FIFO is empty
  
  assign ffull_o = tfifo_full;
  assign fempty_o = tfifo_empty;
   
  always @ (negedge wclk_i)          // Write FIFO logic
    begin
      if (!rst_i)       // Reset condition
        begin
          adrw <= 12'h0;
          wack_o <= 1'b0;
        end
      else if (we_i & ~tfifo_full)
        begin
          tfifo[adrw[10:0]] <= data_i;
          adrw <= adrw + 1'b1;
          wack_o <= 1'b1;
        end
      else
        begin
          wack_o <= 1'b0;
        end
      adrr_s <= adrr;         // Synchronized read pointer
      
    end
    
  always @ (posedge wclk_i)
    begin
      if (!rst_i)
        begin
          tfifo_full <= 1'b0;
          afull_o <= 1'b0;
        end
      else
        begin
          if ({~adrw[11],adrw[10:0]} == adrr_s[11:0])
            tfifo_full <= 1'b1;
          else
            tfifo_full <= 1'b0;
        end
      fifo_count_o <= {5'h0,(adrw[10:0] - adrr_s[10:0])};
      if ((adrw[10:0] - adrr_s[10:0]) > `FIFO_AFULL)
        afull_o <= 1'b1;
      else
        afull_o <= 1'b0;   
    end  
    
  always @ (negedge rclk_i)          // Read FIFO logic
    begin
      if (!rst_i)
        begin
          adrr <= 12'h0;
          rack_o <= 1'b0;
        end
      else if (rd_i & ~tfifo_empty)
        begin
          data_o <= tfifo[adrr[10:0]];
          adrr <= adrr + 1'b1;
          rack_o <= 1'b1;
        end
      else
        begin
          rack_o <= 1'b0;
        end
      adrw_s <= adrw;               // Synchronized write pointer

    end

  always @ (posedge rclk_i)
    begin
      if (!rst_i)
        begin
          tfifo_empty <= 1'b1;
          aempty_o <= 1'b0;      
        end
      else
        begin
          if (adrr == adrw_s)
            tfifo_empty <= 1'b1;
          else
            tfifo_empty <= 1'b0; 
        end
      if ((adrw_s[10:0] - adrr[10:0]) < `FIFO_AEMPTY)
        aempty_o <= 1'b1;
      else
        aempty_o <= 1'b0;
    end    
    


  
endmodule
