//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Gigabit ethernet controller top module
//
`include "parms.v"

module gig_ether (
  // PHY interface
  clk_i, clk125_i, rx_clk_i, tx_clk_i,
  rst_i, intphy_i, phy_crs_i, phy_col_i, rx_er_i, rx_dv_i, rxd_i,
  phy_link10_i, phy_link100_i, phy_link1000_i, phy_duplex_i,
  phy_ledrx_i, phy_ledtx_i,
  gtx_clk_o, mdc_o, rstphy_o, tx_er_o, tx_en_o, txd_o,
  mdio_io,
  // Wishbone interface
  wb_adr_i, wb_data_i, wb_we_i, wb_sel_i, wb_stb_i, wb_ack_i,
  wb_cyc_i, wb_tagn_i,
  wb_adr_o, wb_data_o, wb_we_o, wb_sel_o, wb_stb_o, wb_ack_o,
  wb_cyc_o, wb_tagn_o,
  // Test I/O
  rx_fsm_o, rxmac_fsm_o, tx_fsm_o, txmac_fsm_o, byte_cnt_o
  );
  
  input clk_i;                      // Input clock
  input clk125_i;                   // Input 125MHz clock
  input rx_clk_i;                   // GMII and MII receive clock. 125MHz in 1000BASE-T,
                                    // 25MHz in 100BASE-TX and 2.5MHz in 10BASE-T
  input tx_clk_i;                   // MII transmit clock. 25MHz in 100BASE-TX and 2.5MHz in 10BASE-T
  input rst_i;                      // Input reset signal
  input intphy_i;                   // Interrupt request from PHY
  input phy_crs_i;                  // GMII and MII carrier sense. CRS asserts when the receive medium is
                                    // non-idle.
  input phy_col_i;                  // GMII and MII collision.
  input rx_er_i;                    // GMII and MII Receive Error.
  input rx_dv_i;                    // GMII and MII Receive Data Valid
  input [7:0] rxd_i;                // GMII and MII Receive Data
  input phy_link10_i;               // Speed input from PHY
  input phy_link100_i;              // Speed input from PHY
  input phy_link1000_i;             // Speed input from PHY
  input phy_duplex_i;               // Duplex/collision mode input from PHY                                     
  input phy_ledrx_i;                // Receive activity input from PHY
  input phy_ledtx_i;                // Transmit activity input from PHY
  output gtx_clk_o;                 // GMII 125MHz transmit clock
  output mdc_o;                     // Output clock to PHY
  output rstphy_o;                  // When low PHY is reset
  output tx_er_o;                   // GMII and MII Transmit Error
  output tx_en_o;                   // GMII and MII Transmit Enable
  output [7:0] txd_o;               // GMII and MII Transmit Data
  inout mdio_io;                    // Input/output data from/to  PHY
  
  // Wishbone interface
  input [31:0] wb_adr_i;            // Input address
  input [31:0] wb_data_i;           // Input data
  input wb_we_i;                    // Write enable signal
  input [3:0] wb_sel_i;             // Byte select signal
  input  wb_stb_i;                  // Strobe input - select signal
  input wb_ack_i;                   // Acknowledge signal
  input wb_cyc_i;                   // Bus cycle indicator
  input wb_tagn_i;                  // Tag input signal
  output [31:0] wb_adr_o;           // Output address
  output [31:0] wb_data_o;          // Output data
  output wb_we_o;                   // Output write enable signal
  output [3:0] wb_sel_o;            // Byte select signal
  output wb_stb_o;                  // Strobe output
  output wb_ack_o;                  // Acknowledge signal
  output wb_cyc_o;                  // Bus cycle indicator
  output wb_tagn_o;                 // Tag output signal   
  
  // Test I/O
  output [3:0] rx_fsm_o;            // To monitor state of rx bufmgr FSM
  output [3:0] rxmac_fsm_o;         // To monitor state of rxmac FSM  
  output [3:0] tx_fsm_o;            // States of tx bufmgr FSM
  output [3:0] txmac_fsm_o;         // States of txmac FSM
  output [15:0] byte_cnt_o;         // Byte count from TxMAC FSM
  

 reg rstphy_o;
 // Internal wires and registers
 reg [47:0] mac_adr = 48'h001CBED3DCAD;      // Constant value
 wire [31:0] ops_cr_w;
 wire [31:0] ops_txdp_w;
 wire [31:0] ops_txcfg_w;
 wire [31:0] tx_data_w;
 wire tx_dread_w;
 wire tx_dwrite_w;
 
 wire [31:0] wb_datai_w;         // I/O reference to tx_bufmgr_top
 wire [31:0] wb_datao_w;
 
 wire xfer_done_w;
 wire [15:0] data_count_w;
 wire tfifo_full_w;
 wire data_rq_w;
 wire txdpw_w;
 wire [31:0] int_wbadr_w; 
 wire [31:0] int_wbdata_w;
 wire data_ready_w;
 wire we_opreg_w;
 wire [3:0] wb_fsm_state_w;       // State of wb_cntr FSM
 wire tx_start_w;
 wire [15:0] tx_desc_cnt_w;
 wire tx_fempty_w;
 wire tx_fread_w;
 wire tx_clkf_w;
 wire mac_xfer_w;
 wire mac_failed_w;
 wire [31:0] int_data_w;
 wire rx_ffull_w;
 wire rx_fafull_w;
 wire [31:0] rfcr_w;
 wire [31:0] rxcfg_w;
 wire [31:0] rxdp_w;
 wire [31:0] rxmac_data_w;
 wire rxmac_fwrite_w;
 wire rxdpw_w;
 wire rx_dread_w;
 wire rx_dwrite_w;
 wire [31:0] wb_adrrx_w;
 wire [31:0] wb_datarx_w;
 wire [31:0] int_adr_w;
 wire [31:0] wb_adrtx_w;
 wire [31:0] wb_datatx_w;
 wire [15:0] rx_data_count_w;
 wire [15:0] tx_data_count_w;
 wire [5:0] psize_rd_w;
 wire [5:0] psize_wr_w;
 wire data_valid_w;
 wire [15:0] packet_size_w;
 reg [31:0] wb_data_neg;
 reg [31:0] wb_adr_neg; 
 wire [3:0] test01_w; 
 assign byte_cnt_o = {12'h0,test01_w};
 
 // WB address space decoding

 assign int_wbdata_w = (wb_fsm_state_w == `WB_EXWRITE) ? wb_data_neg :
                       (wb_fsm_state_w == `WB_BREAD) ? int_data_w : 32'hz;
                       
 wire rx_data_access_w = rx_dread_w | rx_dwrite_w;
 wire tx_data_access_w = tx_dread_w | tx_dwrite_w;
 
 wire rd_wish_w = rx_dread_w | tx_dread_w;
 wire wr_wish_w = rx_dwrite_w | tx_dwrite_w;
 
 assign int_adr_w = (rx_data_access_w) ? wb_adrrx_w : wb_adrtx_w;
 assign wb_datao_w = (rx_data_access_w) ? wb_datarx_w : wb_datatx_w;
 
 assign data_count_w = (rx_data_access_w) ? rx_data_count_w : tx_data_count_w;
 
 // Rx data access has higher priority than Tx
 assign rx_ack_w = (rx_data_access_w) ? data_ready_w : 1'b0;
 assign tx_ack_w = (~rx_data_access_w) ? data_ready_w : 1'b0;

// Modules instatiation
  
  rx_bufmgr rx_bufmgr_top (         // Receive buffer manager
    .clk_i                (clk_i),
    .rx_clk_i             (rx_clk_i),
    .rst_i                (rst_i),
    .cr_rxen_i            (ops_cr_w[2]), 
    .wb_data_i            (int_wbdata_w), 
    .rxcfg_i              (rxcfg_w), 
    .ack_i                (rx_ack_w), 
    .rxdpw_i              (rxdpw_w), 
    .rxdp_i               (rxdp_w), 
    .rxmac_data_i         (rxmac_data_w), 
    .fwrite_i             (rxmac_fwrite_w), 
    .xfer_done_i          (xfer_done_w),
    .psize_wr_i           (psize_wr_w),
    .data_valid_i         (data_valid_w),
    .packet_size_i        (packet_size_w),
    .wb_data_o            (wb_datarx_w), 
    .ffull_o              (rx_ffull_w), 
    .fafull_o             (rx_fafull_w), 
    .wb_adr_o             (wb_adrrx_w),
    .rx_dread_o           (rx_dread_w),
    .rx_dwrite_o          (rx_dwrite_w),
    .data_count_o         (rx_data_count_w),
    .psize_rd_o           (psize_rd_w),
    .rx_fsm_o             (rx_fsm_o)
    );
  
  ops_regs ops_regs_top (     // Operational registers
    .clk_i                (clk_i),
    .rst_i                (rst_i),
    .data_i               (int_wbdata_w),
    .adr_i                (wb_adr_neg[7:0]),
    .we_i                 (we_opreg_w),
    .cr_o                 (ops_cr_w),
    .txdp_o               (ops_txdp_w),
    .txcfg_o              (ops_txcfg_w),
    .rxcfg_o              (rxcfg_w),
    .rxdp_o               (rxdp_w),
    .rfcr_o               (rfcr_w),
    .txdpw_o              (txdpw_w),
    .rxdpw_o              (rxdpw_w)
  );
  
 
  rx_mac rx_mac_top (             // Rx MAC function
    .clk_i                (clk_i),
    .rst_i                (rst_i),
    .rx_clk_i             (rx_clk_i), 
    .rx_er_i              (rx_er_i), 
    .rx_dv_i              (rx_dv_i), 
    .rxd_i                (rxd_i), 
    .link_speed_i         ({phy_link1000_i,phy_link100_i,phy_link10_i}),
    .phy_col_i            (phy_col_i), 
    .duplex_i             (phy_duplex_i), 
    .ffull_i              (rx_ffull_w), 
    .affull_i             (rx_fafull_w), 
    .mac_adr_i            (mac_adr), 
    .rfcr_i               (rfcr_w),
    .psize_rd_i           (psize_rd_w),
    .rx_data_o            (rxmac_data_w), 
    .fwrite_o             (rxmac_fwrite_w),
    .psize_wr_o           (psize_wr_w),
    .data_valid_o         (data_valid_w),
    .packet_size_o        (packet_size_w),
    .rxmac_fsm_o          (rxmac_fsm_o)
  
  );
  
  wb_cntr_g wb_cntrg_top (        // Wishbone controller
    .clk_i                (clk_i),
    .rst_i                (rst_i),
    .rd_i                 (rd_wish_w), 
    .int_wr_i             (wr_wish_w),
    .wcntr_i              (data_count_w), 
    .dataex_i             (wb_data_i), 
    .adrex_i              (wb_adr_i), 
    .ack_i                (wb_ack_i),
    .int_adr_i            (int_adr_w),
    .int_data_i           (wb_datao_w),
    .stb_i                (wb_stb_i),
    .cyc_i                (wb_cyc_i),
    .sel_i                (wb_sel_i),
    .we_i                 (wb_we_i),
    .dataex_o             (wb_data_o), 
    .adrex_o              (wb_adr_o), 
    .we_o                 (wb_we_o), 
    .sel_o                (wb_sel_o), 
    .stb_o                (wb_stb_o), 
    .cyc_o                (wb_cyc_o),
    .int_data_o           (int_data_w),
    .xfer_done_o          (xfer_done_w),
    .data_ready_o         (data_ready_w),
    .int_adr_o            (),
    .we_opreg_o           (we_opreg_w),
    .ack_o                (wb_ack_o),
    .fsm_state_o          (wb_fsm_state_w)
  );																								
  
  tx_bufmgr tx_bufmgr_top (
  	.clk_i                (clk_i), 
  	.rst_i                (rst_i), 
  	.xfer_start_i         (wb_cyc_i), 
  	.xfer_done_i          (xfer_done_w), 
  	.cr_txen_i            (ops_cr_w[0]), 
  	.wb_data_i            (int_wbdata_w), 
  	.txcfg_i              (ops_txcfg_w),
    .ack_i                (tx_ack_w), 
    .data_rq_i            (data_rq_w), 
    .txdpw_i              (txdpw_w), 
    .txdp_i               (ops_txdp_w), 
    .tx_fread_i           (tx_fread_w), 
    .tx_clk_i             (tx_clkf_w), 
    .macxfer_done_i       (mac_xfer_w),
    .macfailed_i          (mac_failed_w),
    .tx_data_o            (tx_data_w), 
    .wb_data_o            (wb_datatx_w), 
    .tx_dread_o           (tx_dread_w), 
    .tx_dwrite_o          (tx_dwrite_w),
    .data_count_o         (tx_data_count_w), 
    .tfifo_full_o         (tfifo_full_w), 
    .wb_adr_o             (wb_adrtx_w), 
    .tx_start_o           (tx_start_w), 
    .tx_fempty_o          (tx_fempty_w), 
    .tx_desc_cnt_o        (tx_desc_cnt_w),
    .tx_fsm_o             (tx_fsm_o),
    .test01_o             (test01_w)
  );						
  
  tx_mac tx_mac_top (
  	.clk_i                (clk_i), 
  	.rst_i                (rst_i), 
  	.data_i               (tx_data_w), 
  	.tx_clk_i             (tx_clk_i), 
  	.phy_col_i            (phy_col_i),
    .phy_crs_i            (phy_crs_i), 
    .link_speed_i         ({phy_link1000_i,phy_link100_i,phy_link10_i}), 
    .duplex_i             (phy_duplex_i), 
    .tx_start_i           (tx_start_w),
    .tx_desc_cnt_i        (tx_desc_cnt_w), 
    .tx_fempty_i          (tx_fempty_w),  
    .gtx_clk_i            (clk125_i),
    .data_o               (txd_o), 
    .tx_en_o              (tx_en_o), 
    .tx_er_o              (tx_er_o), 
    .tx_fread_o           (tx_fread_w),
    .tx_clk_o             (tx_clkf_w), 
    .xfer_done_o          (mac_xfer_w), 
    .xfer_failed_o        (mac_failed_w),
    .txmac_fsm_o          (txmac_fsm_o),
    .byte_cnt_o           ()
  );
  
  always @ (posedge clk_i)
    if (!rst_i)
      begin
        rstphy_o <= 1'b0;
      end
    else
      begin
        rstphy_o <= 1'b1;
      end
      
 always @ (negedge clk_i)
   begin
     wb_data_neg <= wb_data_i;
     wb_adr_neg <= wb_adr_i;
   end

  
endmodule