//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Rx buffer manager top file
//
`include "parms.v"

module rx_bufmgr (
  clk_i, rx_clk_i, rst_i, cr_rxen_i, wb_data_i, rxcfg_i, ack_i, rxdpw_i, 
  rxdp_i, rxmac_data_i, fwrite_i, xfer_done_i, psize_wr_i, data_valid_i,
  packet_size_i,
  wb_data_o, ffull_o, fafull_o, wb_adr_o, rx_dread_o, rx_dwrite_o, data_count_o,
  psize_rd_o, rx_fsm_o
  
  );
  input clk_i;                // Input clock
  input rx_clk_i;             // Receiver clock input
  input rst_i;                // Input reset signal
  input cr_rxen_i;            // Rx enable bit from CR operational register
  input [31:0] wb_data_i;     // Input data from Wishbone
  input [31:0] rxcfg_i;       // Rx configuration register input
  input ack_i;                // Signal from wb that data is ready
  input rxdpw_i;              // Input signal to update rxdp, changes on the negedge
  input [31:0] rxdp_i;        // Input Rx descriptor pointer
  input [31:0] rxmac_data_i;  // Input data from MAC
  input fwrite_i;             // FIFO write signal from MAC
  input xfer_done_i;          // Completion of the bus transfer request
  input [5:0] psize_wr_i;     // Write pointer from TxMAC 
  input data_valid_i;         // Signal from TxMAC that data frame is valid
  input [15:0] packet_size_i; // Size of received packet in bytes
  output [31:0] wb_data_o;    // Output data to the wishbone
  output ffull_o;             // FIFO full signal for MAC
  output fafull_o;            // FIFO almost full signal for MAC
  output [31:0] wb_adr_o;     // Address for the wishbone
  output rx_dread_o;          // Requst to read data from the wishbone
  output rx_dwrite_o;         // Request to write data to the wishbone
  output [15:0] data_count_o; // Number of bytes to be read/written from the bus
  output [5:0] psize_rd_o;    // Read  pointer for TxMAC
  output [3:0] rx_fsm_o;      // Test output to monitor state of Rx FSM
  
  reg [15:0] data_count_o;
  reg [31:0] wb_adr_o;
  

  
  // Internal wires and registers
  reg [31:0] rxdp;                      // Rx descriptor pointer										 
  reg [31:0] rxdp_bu;                   // Backup value for RXDP
  reg [31:0] rx_desc_cache_link;        // Cached link of the Rx descriptor
  reg [31:0] rx_desc_cache_bufptr;      // Buffer pointer
  reg [31:0] rx_desc_cache_cmdsts;      // Command/status fields
  reg [31:0] rx_desc_cache_extsts;      // Extended status fields
  
  reg crdd;                             // Flag, set when the current Rx descriptor has been completed
  reg [31:0] rxdp_pos;
  reg rxdpw_pos;
  reg [15:0] desc_cnt;                  // Count of bytes remaining in the current buffer
  reg [15:0] desc_cnt_init;             // Value of desc_cnt at the beginning of the cycle
  reg [15:0] desc_fifo_write;           // Number of dwords 
  reg [15:0] word_count;
  reg [15:0] scroll_count;              // Number of dwords to be scrolled
  reg [15:0] scroll_count_pos;
  reg scroll_done;
  reg scroll_done_pos;
  reg ack_pos;
  reg xfer_done_neg;
  reg rd_r;

  reg [5:0] psize_rd;                  // Pointer to the packet_size array for FIFO data read               
  reg [5:0] psize_wr_s;                // Synchronized pointer
  reg [31:0] rx_data_out;
  reg [31:0] rx_data_out_neg;
  reg [3:0] adv_count;                 // Counter designated for RX_ADVANCE stage
  reg [3:0] adv_count_pos;
  reg [15:0] timer_exp;                // Timer to restrict bus reads during IDLE-DESC read polling
  reg timer_exp_r;	 
  reg [15:0] packet_size;              // Buffer to store packet's size value to be recorded with cmdsts
  
  wire fifo_ready = (psize_rd == psize_wr_s) ? 1'b0 : 1'b1;   // If not equal to zero there is one or more ready packets to be Tx'd
   
  wire [3:0] rx_fsm_w;
  wire [31:0] fifo_datao_w;
  assign wb_data_o =  (rx_fsm_w == `RX_FRAG_WRITE) ? fifo_datao_w:
                      (rx_fsm_w == `RX_ADVANCE) ? rx_data_out_neg : 32'h0;
  
  assign psize_rd_o = psize_rd;
  assign rx_fsm_o = rx_fsm_w;
  
  assign rx_dread = rx_fsm_w[0] && ~rx_fsm_w[1] && ~rx_fsm_w[2] && ~rx_fsm_w[3]; 
  assign rx_fwrite = rx_fsm_w[0] && rx_fsm_w[1] && ~rx_fsm_w[2] && ~rx_fsm_w[3];
  assign rx_advance = ~rx_fsm_w[0] && ~rx_fsm_w[1] && rx_fsm_w[2] && ~rx_fsm_w[3];
  assign rx_dread_o = rx_dread;           // Any command which require data to be read from wb bus
  assign rx_dwrite_o = rx_advance || rx_fwrite;
  
  // Instantiate modules
  rx_fsm rx_fsm_buf (
    .clk_i            (clk_i), 
    .rst_i            (rst_i), 
    .xfer_done_i      (xfer_done_i), 
    .cr_rxen_i        (cr_rxen_i), 
    .crdd_i           (crdd),
    .own_i            (rx_desc_cache_cmdsts[31]),
    .fifo_ready_i     (fifo_ready),
    .data_valid_i     (data_valid_i),
    .scroll_done_i    (scroll_done_pos),
    .timer_exp_i      (timer_exp_r),
    .rx_fsm_o         (rx_fsm_w)
  );
  
  fifo rx_fifo (
    .wclk_i           (rx_clk_i), 
    .rclk_i           (clk_i), 
    .rst_i            (rst_i), 
    .we_i             (fwrite_i), 
    .rd_i             (rd_r), 
    .data_i           (rxmac_data_i),
    .data_o           (fifo_datao_w), 
    .ffull_o          (ffull_o), 
    .fempty_o         (),  
    .fifo_count_o     (), 
    .wack_o           (), 
    .rack_o           (),
    .afull_o          (fafull_o), 
    .aempty_o         () 
  );
  
  always @ (rx_dread or rx_fwrite or rx_advance)
    begin
      if (rx_dread)
        begin
          wb_adr_o = rxdp;
          data_count_o = `RD_DESC;
        end
      else if (rx_fwrite)
        begin
          wb_adr_o = rx_desc_cache_bufptr;
          data_count_o = desc_cnt_init;
        end
      else if (rx_advance)    // Descriptor needs to be written back to the main memory
        begin
          wb_adr_o = rxdp_bu;
          data_count_o = `RD_DESC;
        end
    end
  
  always @ (posedge clk_i)
    begin
      rxdpw_pos <= rxdpw_i;
      rxdp_pos <= rxdp_i;
      ack_pos <= ack_i;
      scroll_done_pos <= scroll_done;
      scroll_count_pos <= scroll_count;
      adv_count_pos <= adv_count;
    end
    
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          psize_rd <= 6'h0;
          psize_wr_s <= 6'h0;
        end
      else
        begin
          psize_wr_s <= psize_wr_i;             // Synchronization of write pointers
          if (xfer_done_neg && (rx_fsm_w == `RX_FRAG_WRITE))
            psize_rd <= psize_rd + 1'b1; 
          else if (scroll_done)
            psize_rd <= psize_rd + 1'b1;       
        end
    end
    
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          rd_r <= 1'b0;  
          timer_exp <= 16'h0;
          timer_exp_r <= 1'b0;
        end
      else
        begin
          case (rx_fsm_w)                 // Synthesis full case
            `RX_IDLE:
              begin
                if (timer_exp == `RX_TIMER_EXP)
                  begin
                    timer_exp_r <= 1'b1;
                  end
                else
                  begin
                    timer_exp <= timer_exp + 1'b1;
                    timer_exp_r <= 1'b0;
                  end
              end
            `RX_DESC_READ:
              begin
                timer_exp <= 16'h0;
                timer_exp_r <= 1'b0;
              end
            `RX_FRAG_WRITE:
              begin
                if (word_count == 16'h1)
                  begin
                    rd_r <= 1'b1;
                    packet_size <= packet_size_i;
                  end
                else
                  if (ack_i)
                    rd_r <= 1'b1;
                  else
                    rd_r <= 1'b0;
              end
            `RX_SCROLL:
              begin
                rd_r <= 1'b1;
              end
            `RX_FIFO_BLOCK:
              begin
                rd_r <= 1'b0;																		
              end
            `RX_ADVANCE:
              begin
                timer_exp <= `RX_TIMER_EXP;
                if (~ack_i)
                  rx_data_out <= rx_desc_cache_link;
                else
                  begin
                    case (adv_count)                // Synthesis full case
                      4'h0: rx_data_out <= rx_desc_cache_bufptr;
                      4'h1: rx_data_out <= {(rx_desc_cache_cmdsts[31:16] | 16'h8800),packet_size};           // Set OWN and 'packet OK' bits and update packet size field  
                      4'h2: rx_data_out <= rx_desc_cache_extsts;
                    endcase
                  end
              end
          endcase
        end
    end
  
  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          crdd <= 1'b1;
        end
      else
        begin
          if (rxdpw_pos)
            begin
              crdd <= 1'b0;         
              rxdp <= rxdp_pos;       // New RXDP loaded by the driver
            end
          else if ((adv_count_pos == 4'h3) && (rx_fsm_w == `RX_ADVANCE) && (rx_desc_cache_link == 32'h0))
            crdd <= 1'b1;       // Current RXDP has been completed
        end
      case (rx_fsm_w)               // Synthesis full case
        `RX_IDLE:
          begin
            word_count <= 16'h0;
          end
        `RX_DESC_READ:
          begin													
            rxdp_bu <= rxdp;
            case ({word_count,ack_pos})           // Synthesis full case
              {16'h0,1'b1}: rx_desc_cache_link <= wb_data_i;
              {16'h1,1'b1}: rx_desc_cache_bufptr <= wb_data_i;
              {16'h2,1'b1}: rx_desc_cache_cmdsts <= wb_data_i;
              {16'h3,1'b1}: rx_desc_cache_extsts <= wb_data_i;
            endcase  
            if (xfer_done_i)
              begin
                desc_cnt <= rx_desc_cache_cmdsts[15:0];
                desc_fifo_write <= {2'b00,rx_desc_cache_cmdsts[15:2]};      // Number of dwords
              end
            if (ack_pos)
              word_count <= word_count + 1'b1;
          end
        `RX_FIFO_BLOCK:
          begin
            word_count <= 16'h0;
            if (packet_size_i[1:0] === 2'b00)
              begin
                scroll_count <= {2'b00,packet_size_i[15:2]};
                desc_cnt_init <= packet_size_i;
              end
            else
              begin
                scroll_count <= {2'b00,packet_size_i[15:2]} + 1'b1;
                desc_cnt_init <= {packet_size_i[15:2],2'b00} + 3'h4;
              end
            scroll_done <= 1'b0;
          end
        `RX_FRAG_WRITE:       // Need to write one frame from FIFO to the main memory
          begin
            word_count <= word_count + 1'b1;
            adv_count <= 4'h0;
          end
        `RX_ADVANCE:
          begin
            if (ack_pos)
              adv_count <= adv_count + 1'b1;
            if (rx_desc_cache_link != 32'h0)
              rxdp <= rx_desc_cache_link;
          end
        `RX_SCROLL:
          begin
            scroll_count <= scroll_count - 1'b1;
            if (scroll_count == 16'h2)
              scroll_done <= 1'b1;
            else
              scroll_done <= 1'b0;
          end
      endcase
    end
    
  always @ (negedge clk_i)
    begin
      xfer_done_neg <= xfer_done_i;
      rx_data_out_neg <= rx_data_out;
    end
  
endmodule
