//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// RX state machine
// 
`include "parms.v"

module rx_fsm (
  clk_i, rst_i, xfer_done_i, cr_rxen_i, crdd_i, own_i, fifo_ready_i,
  data_valid_i, scroll_done_i, timer_exp_i,
  rx_fsm_o
  
  );
  
  input clk_i;                    // Input clock signal
  input rst_i;                    // Input reset signal
  input xfer_done_i;              // Completion of a PCI bus transfer request
  input cr_rxen_i;                // RX Enable bit from CR operational register
  input crdd_i;                   // Input crdd from rx_bufmgr
  input own_i;                    // Descriptor OWN bit
  input fifo_ready_i;             // When set FIFO has completed packets
  input data_valid_i;             // Signal from RxMAC that valid packet is received
  input scroll_done_i;            // Signal that scroll of FIFO is done
  input timer_exp_i;              // Timer to allow to go to RX_DESC_READ state
  output [3:0] rx_fsm_o;          // FSM states output
  
  reg [3:0] rx_fsm_state;       
  reg [3:0] rx_fsm_next_state;
  
  // Internal wires and registers
  assign rx_fsm_o = rx_fsm_state;
  
  always @ (rx_fsm_state or cr_rxen_i or crdd_i or xfer_done_i or own_i    // All commands asserted on the posedge
            or fifo_ready_i or scroll_done_i or timer_exp_i)
    begin
      rx_fsm_next_state = `RX_IDLE;
      
      case (rx_fsm_state)           // Synthesis full case
        `RX_IDLE:
          begin
            if (cr_rxen_i && ~crdd_i && timer_exp_i)
              rx_fsm_next_state = `RX_DESC_READ;
            else
              rx_fsm_next_state = `RX_IDLE;
          end
        `RX_DESC_READ:
          begin
            if (xfer_done_i && ~own_i)
              rx_fsm_next_state = `RX_FIFO_BLOCK;
            else if (xfer_done_i && own_i)
              rx_fsm_next_state = `RX_IDLE;
            else
              rx_fsm_next_state = `RX_DESC_READ;        // Otherwise stay in the same position
          end
        `RX_FIFO_BLOCK:
          begin
            if (fifo_ready_i && data_valid_i)       // Need to write one packet of data to the main memory
              rx_fsm_next_state = `RX_FRAG_WRITE;           
            else if (fifo_ready_i && ~data_valid_i)
              rx_fsm_next_state = `RX_SCROLL;       // There is a packet in the memory but data is not valid. Just scroll it
            else
              rx_fsm_next_state = `RX_FIFO_BLOCK;       // Stay in the same position
          end
        `RX_FRAG_WRITE:
          begin
            if (xfer_done_i)
              rx_fsm_next_state = `RX_ADVANCE;
            else
              rx_fsm_next_state = `RX_FRAG_WRITE;             // Stay in the same position
          end
        `RX_ADVANCE:
          begin
            if (xfer_done_i)
              rx_fsm_next_state = `RX_IDLE;
            else
              rx_fsm_next_state = `RX_ADVANCE;
          end
        `RX_SCROLL:
          begin
            if (scroll_done_i)
              rx_fsm_next_state = `RX_FIFO_BLOCK;
            else
              rx_fsm_next_state = `RX_SCROLL;
          end
      endcase
    end
    
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          rx_fsm_state <= `RX_IDLE;
        end
      else
        begin
          rx_fsm_state <= rx_fsm_next_state;
        end
    end
    
  
endmodule