//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
`include "parms.v"
  
module rx_mac (
  clk_i, rst_i, rx_clk_i, rx_er_i, rx_dv_i, rxd_i, link_speed_i,
  phy_col_i, duplex_i, ffull_i, affull_i, mac_adr_i, rfcr_i, psize_rd_i,
  rx_data_o, fwrite_o, psize_wr_o, data_valid_o, packet_size_o, rxmac_fsm_o
  );
  
  input clk_i;                      // Input clock
  input rst_i;                      // Input reset signal
  input rx_clk_i;                   // Input Rx clock from PHY
  input rx_er_i;                    // Receive error
  input rx_dv_i;                    // Data valid
  input [7:0] rxd_i;                // Input data
  input [2:0] link_speed_i;         // Speed of the link
  input phy_col_i;                  // Collision
  input duplex_i;                   // Duplex mode input from PHY
  input ffull_i;                    // FIFO full input signal - asserted on the posedge
  input affull_i;                   // FIFO Almost full input signal
  input [47:0] mac_adr_i;           // MAC address
  input [31:0] rfcr_i;              // Filtering control register, bit 31 is RFEN
  input [5:0] psize_rd_i;           // Read pointer from Tx bufmgr
  output [31:0] rx_data_o;          // Output data for rx_bufmgr
  output fwrite_o;                  // FIFO write signal
  output [5:0] psize_wr_o;          // Write pointer for Tx bufmgr
  output data_valid_o;              // Signal to Tx bufmgr that valid data has been acquired
  output [15:0] packet_size_o;      // Size of the received packet in bytes	 
  output [3:0] rxmac_fsm_o;         // Test output - to monitor state of rxmac FSM
  
  reg [31:0] rx_data_o;
  reg fwrite_o;
  reg data_valid_o;
  reg [15:0] packet_size_o;
  
  // Internal wires and registers
  
  wire rx_clk = rx_clk_i;
  reg [31:0] rx_data;         // Register to hold data until it's sent to the bufmgr
  reg [31:0] rx_data_prev;
  reg [47:0] mac_adr;         // Register to store destination MAC address for the received frame
  reg [15:0] packet_size [0:31];    // Buffer to keep sizes of received data packets. Maximum 32 packets are allowed
  reg data_valid [0:31];            // Array of data valid flags
  reg [5:0] psize_rd_s;             // Sychronized read pointer
  reg [5:0] psize_wr;               // Pointer to the packet_size array for FIFO data write
  reg pbuf_full;
  
  reg [15:0] byte_cnt;              // Received bytes counter
  reg rxc;                          // Confirmation that packet to be received and not filtered out
  reg [1:0] count;                  // Register to count bytes in dword
  reg frame_done;                   // Reception of the frame including padding (if exists) completed
  
  reg crc_init_r;
  reg crc_dataen_r;
  reg crc_rd_r;
  wire [7:0] crco_w;
  reg [31:0] crc_value;             // Calculated CRC value
  reg [31:0] crc_val_acq;           // Acquired CRC value
  reg [7:0] rxd_crc_pos; 
  reg next_state1;
  reg next_state2;
  
  reg [7:0] rxd1_r;                 // Set of registers to delay data four clocks for CRC calculation
  reg [7:0] rxd2_r;
  reg [7:0] rxd3_r;
  reg [7:0] rxd4_r;
  reg [7:0] rxd5_r;	  
  reg [7:0] rxd6_r;
  
  assign psize_wr_o = psize_wr;
  
  // Modules instantiation
    crc_gen crc_gen_rxmac (
    .clk_i                (rx_clk), 
    .rst_i                (rst_i),
    .init_i               (crc_init_r), 
    .frame_data_i         (rxd6_r), 
    .data_en_i            (crc_dataen_r), 
    .crc_rd_i             (crc_rd_r),
    .crc_o                (crco_w)
  );
  
  //---------------------------RxMAC FSM definition------------------------
  reg [3:0] rxmac_fsm_state;
  reg [3:0] rxmac_fsm_next_state;				
  assign rxmac_fsm_o = rxmac_fsm_state;
                                                   
  always @ (rxmac_fsm_state or rx_er_i or rx_dv_i or next_state1 or next_state2 or affull_i    // All signals in the sensitivity list must be changing on the posedge
            or rxc or pbuf_full)
    begin                                                                        
      rxmac_fsm_next_state = `RXMAC_IDLE;
      
      case (rxmac_fsm_state)                // Synthesis full case
        `RXMAC_IDLE:
          begin
            if (rx_dv_i && ~rx_er_i)
              rxmac_fsm_next_state = `RXMAC_START;
            else
              rxmac_fsm_next_state = `RXMAC_IDLE;
          end
        `RXMAC_START:         // Receive first six bytes of the packet - to check destination address
          begin
            if (rx_dv_i)
              begin
                if (affull_i | pbuf_full )       // FIFO is almost full - packet is dropped
                  rxmac_fsm_next_state = `RXMAC_FFULL;
                else if (next_state1 && rxc)
                  rxmac_fsm_next_state = `RXMAC_FDDATA;
                else if (next_state1 && ~rxc)
                  rxmac_fsm_next_state = `RXMAC_FILTER;
                else
                rxmac_fsm_next_state = `RXMAC_START;          // Stay in the same position
              end
            else
              begin
                rxmac_fsm_next_state = `RXMAC_DVL;      // If rx_dv_i dropped to zero during this stage - error occured
              end        
          end
        `RXMAC_FILTER:          // Packet to be discarded
          begin
            if (~rx_dv_i)
              rxmac_fsm_next_state = `RXMAC_IDLE;           // Wait until packet is Tx'd, but don't receive it
            else 
              rxmac_fsm_next_state = `RXMAC_FILTER;
          end
        `RXMAC_FDDATA:
          begin												 
            if (~rx_dv_i)
              rxmac_fsm_next_state = `RXMAC_FDDC;
            else
              rxmac_fsm_next_state = `RXMAC_FDDATA;        
          end
        `RXMAC_FDDC:
          begin
            rxmac_fsm_next_state = `RXMAC_CRC;            // This state always takes only one clock cycle
          end
        `RXMAC_CRC:
          begin											
            if (next_state1)
              rxmac_fsm_next_state = `RXMAC_IDLE;           // Transmission is completed
            else    
              rxmac_fsm_next_state = `RXMAC_CRC;   
          end
        `RXMAC_FFULL:
          begin
            if (~rx_dv_i)
              rxmac_fsm_next_state = `RXMAC_IDLE;
            else
              rxmac_fsm_next_state = `RXMAC_FFULL;
          end
        `RXMAC_DVL:
          begin               // Just go back to the idle state
            rxmac_fsm_next_state = `RXMAC_IDLE; 
          end
      endcase
    end
    
  always @ (negedge rx_clk)
    begin
      if (!rst_i)
        begin
          rxmac_fsm_state <= `RXMAC_IDLE;
        end
      else
        begin
          rxmac_fsm_state <= rxmac_fsm_next_state;
        end
    end
  
  //-------------------------End of RxMAC FSM definition-------------------
  
  always @ (posedge rx_clk)
    begin
      if (psize_rd_s == {~psize_wr[5],psize_wr[4:0]})
        pbuf_full <= 1'b1;
      else
        pbuf_full <= 1'b0;
    end
    
  always @ (posedge rx_clk)
    begin
      if (!rst_i)
        begin
          psize_rd_s <= 6'h0;
        end
      else
        begin
          psize_rd_s <= psize_rd_i;
        end
    end
    
  always @ (negedge rx_clk)
    begin
      packet_size_o <= packet_size[psize_rd_s[4:0]];
      data_valid_o <= data_valid[psize_rd_s[4:0]];
    end
  
  always @ (posedge rx_clk)
    begin
      if (!rst_i)
        begin
          byte_cnt <= 16'h0;
          fwrite_o <= 1'b0;
          frame_done <= 1'b0;
          crc_dataen_r <= 1'b0;
          crc_rd_r <= 1'b0;
          psize_wr <= 6'h0; 
        end
      else
        begin
          if (rxmac_fsm_state == `RXMAC_IDLE)
            begin
              frame_done <= 1'b0;
              crc_init_r <= 1'b1;
              next_state1 <= 1'b0;													   
              fwrite_o <= 1'b0;                // Test line
              if (rx_dv_i)
                begin
                  rx_data <= {rx_data[23:0],rxd_i};
                  byte_cnt <= byte_cnt + 1'b1;
                  count <= count + 1'b1;
                end
              else
                begin
                  byte_cnt <= 16'h0;
                  count <= 2'b00;
                end
            end
          else if (rxmac_fsm_state == `RXMAC_START)     // Need to acquire preamble,SFD and destination address
            begin
              // Preamble and SFD - first 8 bytes, not sent to FIFO
              byte_cnt <= byte_cnt + 1'b1;
              count <= count + 1'b1;
              if ((byte_cnt > 16'h7) && (byte_cnt < 16'hE))          // First eight bytes not acquired        16'h7
                begin
                  mac_adr <= {mac_adr[39:0],rxd_i};
                end
              if (byte_cnt == 16'hE)                // 16'hE
                begin
                  rx_data <= {8'h00,mac_adr[15:0],rxd_i};
                  if ((~rfcr_i[31]) || (rfcr_i[31] && (mac_adr == mac_adr_i)))            // Filtering is not enabled or dest addr matches
                    begin
                      rxc <= 1'b1;               // Confirm futher reception of the frame
                      rx_data_o <= mac_adr[47:16];
                      fwrite_o <= 1'b1;          // Write data to FIFO
                    end
                  else
                    rxc <= 1'b0;               // Filter this packet out
                  next_state1 <= 1'b1;        // Go to the next stage of FSM
                end
              if (byte_cnt == 16'hD)                // 16'hC
                begin
                  crc_init_r <= 1'b0;
                  crc_dataen_r <= 1'b1;	  
                end
//              fwrite_o <= 1'b1;                                    // Test line
//              rx_data_o <= {rxd6_r,rxd_i, 7'b0, crc_dataen_r,byte_cnt[7:0]};                // Test line
            end 
          else if (rxmac_fsm_state == `RXMAC_FILTER)        // Packet to be fitered out
            begin
              
            end
          else if (rxmac_fsm_state == `RXMAC_FDDATA)      // Receive packet
            begin	
              count <= count + 1'b1;
              byte_cnt <= byte_cnt + 1'b1;
              rx_data <= {rx_data[23:0],rxd_i};
              rx_data_prev <= rx_data;
//              fwrite_o <= 1'b1;                            // Test line
//              rx_data_o <= {rxd6_r,rxd_i, 7'b0, crc_dataen_r, byte_cnt[7:0]};        // Test line
              if (count == 2'b00)                   // 2'b00
                begin
                  rx_data_o <= rx_data;
                  fwrite_o <= 1'b1;             // Write to FIFO
                end
              else
                fwrite_o <= 1'b0;                              
            end            
          else if (rxmac_fsm_state == `RXMAC_FDDC)
            begin						 
              if (count != 2'b01)
                fwrite_o <= 1'b1;
              else
                fwrite_o <= 1'b0;
              rx_data_o <= rx_data_prev;        
//              rx_data_o <= {rxd6_r,rxd_i,7'b0,crc_dataen_r,byte_cnt[7:0]};            // Test line
              crc_val_acq <= rx_data_prev;          // The last four bytes is always CRC
              crc_dataen_r <= 1'b0;
              crc_rd_r <= 1'b1;
              count <= 2'b00;
              crc_value <= {crc_value[23:0],crco_w};
              next_state1 <= 1'b0;
            end
          else if (rxmac_fsm_state == `RXMAC_CRC)   // Need to calculate CRC
            begin
              fwrite_o <= 1'b0;	 												 
              count <= count + 1'b1;
              if (count != 2'b11)          
                begin                                    
                  crc_value <= {crc_value[23:0],crco_w};       
  //                rx_data_o <= crc_val_acq;            // Test line
                end                                                    
              else
                begin
                  if (crc_value != crc_val_acq)       // CRC error
                    data_valid[psize_wr[4:0]] <= 1'b0;
                  else                                // CRC is correct
                    data_valid[psize_wr[4:0]] <= 1'b1;         
//                  packet_size[psize_wr[4:0]] <= 16'h3FF;            // Test line
                  packet_size[psize_wr[4:0]] <= (byte_cnt - 16'h9);     // Preamble and SFD (8 bytes) are counted but not sent to FIFO
                  psize_wr <= psize_wr + 1'b1;
                  next_state1 <= 1'b1;                // Next stage of FSM
//                  rx_data_o <= crc_value;            // Test line
//                  fwrite_o <= 1'b1;                   // Test line
                end
            end                     
          else if (rxmac_fsm_state == `RXMAC_DVL)
            begin
              fwrite_o <= 1'b0;   // To make sure to deassert FIFO write signal after exit from RXMAC_FDSTART stage 
            end
        end 
      rxd1_r <= rxd_i;
      rxd2_r <= rxd1_r;
      rxd3_r <= rxd2_r;
      rxd4_r <= rxd3_r;
      rxd5_r <= rxd4_r;
      rxd6_r <= rxd5_r;
    end
    
endmodule
