//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
module sopc (
  clk_pos_i, clk_neg_i, phy_crs_i, rx_clk_i, phy_col_i, 
  rx_er_i, rx_dv_i, rxd_i, btn_i, sin_i, 
  gtx_clk_o, txd_o, tx_en_o, tx_er_o, led_o, sout_o, mdc_phy_o,
  phy_reset_o
  );
  
  input clk_pos_i;                // Positive differential clock input
  input clk_neg_i;                // Negative differential clock input
  input phy_crs_i;                // Carrier sense input from PHY
  input rx_clk_i;                 // Input Rx clock from PHY
  input phy_col_i;                // Collision detection
  input rx_er_i;                  // Receive error signal from PHY
  input rx_dv_i;                  // Data valid
  input [7:0] rxd_i;              // Data input
  input [3:0] btn_i;              // Input buttons from the board
  input sin_i;                    // UART input
  output gtx_clk_o;               // Output Tx clock for PHY in 1G mode
  output [7:0] txd_o;             // Output Tx data from MAC controller to PHY
  output tx_en_o;                 // Tx enable signal to PHY
  output tx_er_o;                 // Tx error signal to PHY
  output [3:0] led_o;             // LED outputs
  output sout_o;                  // UART output
  output mdc_phy_o;               // Management clock for PHY
  output phy_reset_o;             // Signal to reset PHY - active low
  
  // Internal registers and wires
  wire [31:0] wb_adri_w;          // I/O directions reference to the gig_ether_top 
  wire [31:0] wb_datai_w; 
  wire wb_wei_w; 
  wire [3:0] wb_seli_w; 
  wire wb_stbi_w; 
  wire wb_acki_w;
  wire wb_cyci_w; 
  wire wb_tagni_w;
  wire [31:0] wb_adro_w; 
  wire [31:0] wb_datao_w; 
  wire wb_weo_w; 
  wire [3:0] wb_selo_w; 
  wire wb_stbo_w; 
  wire wb_acko_w;
  wire wb_cyco_w; 
  wire wb_tagno_w;

  wire [31:0] wbcp_adr_w; 
  wire [31:0] wbcp_data_w; 
  wire wbcp_we_w; 
  wire [3:0]  wbcp_sel_w; 
  wire wbcp_stb_w; 
  wire wbcp_ack_w;
  wire wbcp_tagn_w;
  wire wbm_stb_w;
  wire [31:0] wbm_data_w;
  wire wbm_cyc_w;
  wire [31:0] wbm_datao_w;
  wire gtx_clk;
  wire gp_clk;			   
  wire uart_rxrdy_w;
  wire uart_txrdy_w;
  wire [7:0] uart_datar_w;
  wire uart_rd_w;
  wire [7:0] uart_data_w;
  wire wbcp_cyc_w; 
  wire [3:0] rx_fsm_w;
  wire [3:0] rxmac_fsm_w;			
  wire [3:0] tx_fsm_w;
  wire [3:0] txmac_fsm_w;					
  wire [15:0] byte_cnt_w;
 
  reg mdc_phy_o = 1'bz;				   
  
  reg [2:0] linkspeed_r = 3'b100;
  reg duplex_r = 1'b1;	
  reg [2:0] uart_adr_r = 3'b000;
  reg uart_cs_r = 1'b1;	 
  reg [31:0] gp_counter; 
  reg init_done;							 
  reg [15:0] init_counter = 16'h0;
 
  
  wire rst_w = (init_done == 1'b0) ? 1'b0 : ~btn_i[0]; 
  assign phy_reset_o = ~btn_i[1];
  assign led_o = (~btn_i[3]) ? tx_fsm_w : txmac_fsm_w; 
  
   clk_wiz_v1_7_0 gp_clocks
   (// Clock in ports
    .CLK_IN1_P          (clk_pos_i),    // IN
    .CLK_IN1_N          (clk_neg_i),    // IN
    // Clock out ports
    .CLK_OUT1           (gtx_clk),     // OUT
    .CLK_OUT2           (gp_clk)    // OUT
    );
	
//      ODDR2     : In order to incorporate this function into the design,
//     Verilog   : the following instance declaration needs to be placed
//    instance   : in the body of the design code.  The instance name
//   declaration : (ODDR2_inst) and/or the port declarations within the
//      code     : parenthesis may be changed to properly reference and
//               : connect this function to the design.  Delete or comment
//               : out inputs/outs that are not necessary.

//  <-----Cut code below this line---->

   // ODDR2: Output Double Data Rate Output Register with Set, Reset
   //        and Clock Enable.
   //        Spartan-6
   // Xilinx HDL Language Template, version 12.3

   ODDR2 #(
      .DDR_ALIGNMENT("NONE"), // Sets output alignment to "NONE", "C0" or "C1" 
      .INIT(1'b0)    // Sets initial state of the Q output to 1'b0 or 1'b1
//      .SRTYPE("SYNC") // Specifies "SYNC" or "ASYNC" set/reset
   ) clock_fwd (
      .Q(gtx_clk_o),   // 1-bit DDR output data
      .C0(gtx_clk),   // 1-bit clock input
      .C1(~gtx_clk),   // 1-bit clock input
      .CE(1'b1), // 1-bit clock enable input
      .D0(1'b1), // 1-bit data input (associated with C0)
      .D1(1'b0) // 1-bit data input (associated with C1)
 //     .R(1'b0),   // 1-bit reset input
 //     .S(1'b0)    // 1-bit set input
   );

   // End of ODDR2_inst instantiation
					
  // Modules instantiation
  gig_ether gig_ether_top (
    .clk_i                      (gp_clk), 
    .clk125_i                   (gtx_clk), 
    .rx_clk_i                   (rx_clk_i), 
    .tx_clk_i                   (),
    .rst_i                      (rst_w), 
    .intphy_i                   (), 
    .phy_crs_i                  (phy_crs_i), 
    .phy_col_i                  (phy_col_i), 
    .rx_er_i                    (rx_er_i), 
    .rx_dv_i                    (rx_dv_i), 
    .rxd_i                      (rxd_i),   
    .phy_link10_i               (linkspeed_r[0]), 
    .phy_link100_i              (linkspeed_r[1]), 
    .phy_link1000_i             (linkspeed_r[2]), 
    .phy_duplex_i               (duplex_r),
    .phy_ledrx_i                (), 
    .phy_ledtx_i                (), 
    .gtx_clk_o                  (), 
    .mdc_o                      (), 
    .rstphy_o                   (), 
    .tx_er_o                    (tx_er_o), 
    .tx_en_o                    (tx_en_o), 
    .txd_o                      (txd_o),
    .mdio_io                    (),
    .wb_adr_i                   (wb_adri_w), 
    .wb_data_i                  (wb_datai_w), 
    .wb_we_i                    (wb_wei_w), 
    .wb_sel_i                   (wb_seli_w), 
    .wb_stb_i                   (wb_stbi_w), 
    .wb_ack_i                   (wb_acki_w),
    .wb_cyc_i                   (wb_cyci_w), 
    .wb_tagn_i                  (wb_tagni_w),
    .wb_adr_o                   (wb_adro_w), 
    .wb_data_o                  (wb_datao_w), 
    .wb_we_o                    (wb_weo_w), 
    .wb_sel_o                   (wb_selo_w), 
    .wb_stb_o                   (wb_stbo_w), 
    .wb_ack_o                   (wb_acko_w),
    .wb_cyc_o                   (wb_cyco_w), 
    .wb_tagn_o                  (wb_tagno_w),
    .rx_fsm_o                   (rx_fsm_w),
    .rxmac_fsm_o                (rxmac_fsm_w),
    .tx_fsm_o                   (tx_fsm_w),
    .txmac_fsm_o                (txmac_fsm_w),
    .byte_cnt_o                 (byte_cnt_w)
    );
      
    
  wb_host wb_host_top (
    .clk_i                      (gp_clk),
    .rst_i                      (rst_w),
    .wb_adr_i                   (wb_adro_w), 
    .wb_data_i                  (wb_datao_w), 
    .wb_we_i                    (wb_weo_w), 
    .wb_sel_i                   (wb_selo_w), 
    .wb_stb_i                   (wb_stbo_w), 
    .wb_ack_i                   (wb_acko_w),
    .wb_cyc_i                   (wb_cyco_w), 
    .wb_tagn_i                  (wb_tagno_w),
    .wb_adr_o                   (wb_adri_w), 
    .wb_data_o                  (wb_datai_w), 
    .wb_we_o                    (wb_wei_w), 
    .wb_sel_o                   (wb_seli_w), 
    .wb_stb_o                   (wb_stbi_w), 
    .wb_ack_o                   (wb_acki_w),
    .wb_cyc_o                   (wb_cyci_w), 
    .wb_tagn_o                  (wb_tagni_w),
    .wbcp_adr_i                 (wbcp_adr_w), 
    .wbcp_data_i                (wbcp_data_w), 
    .wbcp_we_i                  (wbcp_we_w), 
    .wbcp_sel_i                 (wbcp_sel_w), 
    .wbcp_stb_i                 (wbcp_stb_w), 
    .wbcp_ack_i                 (wbcp_ack_w),
    .wbcp_cyc_i                 (wbcp_cyc_w), 
    .wbcp_tagn_i                (wbcp_tagn_w),
    .wbm_data_i                 (wbm_data_w),
    .wbm_ack_i                  (wbm_ack_w),
    .wbm_cyc_o                  (wbm_cyc_w),
    .wbm_data_o                 (wbm_datao_w)
    );
    
  mem_ctrl mem_ctrl_top (
    .clk_i                      (gp_clk),
    .rst_i                      (rst_w),
    .wb_adr_i                   (wb_adri_w), 
    .wb_data_i                  (wbm_datao_w), 
    .wb_we_i                    (wb_wei_w), 
    .wb_sel_i                   (wb_seli_w), 
    .wb_stb_i                   (wb_stbi_w), 
    .wb_cyc_i                   (wbm_cyc_w), 
    .wb_data_o                  (wbm_data_w), 
    .wb_ack_o                   (wbm_ack_w)
    );  								  
  
  uart_top uart_top_sopc (
  	.clk_i                      (gp_clk),
  	.rst_i                      (rst_w),
  	.sin_i                      (sin_i),
  	.ctsn_i                     (),
  	.sout_o                     (sout_o),
  	.rtsn_o                     (),
  	.data_i                     (uart_data_w),
  	.rd_i                       (uart_rd_w),
  	.wr_i                       (uart_wr_w),
  	.adr_i                      (uart_adr_r),
  	.cs_i                       (uart_cs_r),
  	.data_o                     (uart_datar_w),
  	.irq_o                      (),
  	.txrdy_o                    (uart_txrdy_w),
  	.rxrdy_o                    (uart_rxrdy_w)  	 	
    ); 	
    
  uart_cntr uart_cntr_sopc (
  	.clk_i                      (gp_clk), 
  	.rst_i                      (rst_w), 
  	.data_i                     (wb_datai_w), 
  	.uart_data_i                (uart_datar_w), 		
  	.uart_txrdy_i               (uart_txrdy_w),
  	.uart_rxrdy_i               (uart_rxrdy_w),
  	.uart_we_o                  (uart_wr_w), 
  	.uart_rd_o                  (uart_rd_w),
  	.data_o                     (uart_data_w),
  	.wb_data_i                  (wb_datai_w),
  	.wb_adr_i                   (wb_adri_w),
  	.wb_we_i                    (wb_wei_w),
  	.wb_sel_i                   (wb_seli_w),
  	.wb_stb_i                   (wb_stbi_w),
  	.wb_ack_i                   (wb_acki_w),
  	.wb_cyc_i                   (wb_cyci_w),
  	.wb_data_o                  (wbcp_data_w),
  	.wb_adr_o                   (wbcp_adr_w),
  	.wb_we_o                    (wbcp_we_w),
  	.wb_sel_o                   (wbcp_sel_w),
  	.wb_stb_o                   (wbcp_stb_w),
  	.wb_ack_o                   (wbcp_ack_w),
  	.wb_cyc_o                   (wbcp_cyc_w),
  	.uart_led_o                 ()
  );

  
  always @ (posedge gp_clk)
    begin								  
      if (init_counter != 16'hFFF)
        begin
          init_counter <= init_counter + 1'b1;
          init_done <= 1'b0;
        end
      else					
        init_done <= 1'b1;
    end
 
    
endmodule