//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Top file for Tx data buffer management
//
`include "parms.v"

module tx_bufmgr (
  clk_i, rst_i, xfer_start_i, xfer_done_i, cr_txen_i, wb_data_i, txcfg_i,
  ack_i, data_rq_i, txdpw_i, txdp_i, tx_fread_i, tx_clk_i, macxfer_done_i,
  macfailed_i,
  tx_data_o, wb_data_o, tx_dread_o, tx_dwrite_o,
  data_count_o, tfifo_full_o, wb_adr_o, tx_start_o, tx_fempty_o, tx_desc_cnt_o,
  tx_fsm_o, test01_o
  );
  
  input clk_i;                        // Input clock signal
  input rst_i;                        // Input reset signal
  input xfer_start_i;                 // Start of  bus data transfer
  input xfer_done_i;                  // Completion of a  bus transfer request
  input cr_txen_i;                    // TX enable bit from CR operational register, changes on the negedge
  input [31:0] wb_data_i;             // Input data from the Wishbone bus
  input [31:0] txcfg_i;               // Tx configuration register 
  input ack_i;                        // Signal from wb that data is ready
  input data_rq_i;                    // Data needs to be sent to the bus
  input txdpw_i;                      // Input signal to update txdp, changes on the negedge
  input [31:0] txdp_i;                // Input Tx descriptor pointer
  input tx_fread_i;                   // Signal from Tx MAC to read FIFO
  input tx_clk_i;                     // Read clock for Tx FIFO 
  input macxfer_done_i;               // One packet of data successfully transmitted by MAC
  input macfailed_i;                  // Transmission of one packet failed             
  output [31:0] tx_data_o;            // Output data to Tx MAC
  output [31:0] wb_data_o;            // Output data for the Wishbone bus
  output tx_dread_o;                  // Request to read data from memory
  output tx_dwrite_o;                 // Request to write data to the memory
  output [15:0] data_count_o;         // Number of bytes to be read/written from bus
  output tfifo_full_o;                // Flag that FIFO is full
  output [31:0] wb_adr_o;             // Output address for the Wishbone
  output tx_start_o;                  // Signal to Tx MAC that data FIFO has sufficient data to be Tx'd
  output tx_fempty_o;                 // Output signal to Tx MAC that Tx FIFO is empty
  output [15:0] tx_desc_cnt_o;        // Number of bytes in the current descriptor for Tx MAC
  output [3:0] tx_fsm_o;              // Output for monitoring purposes	  
  output [3:0] test01_o;              // Output register for monitoring purposes
  
  reg [31:0] wb_data_o;
  reg [15:0] data_count_o;
  reg [31:0] wb_adr_o;
  reg tx_start_o;
  
  // Internal registers and wires
  reg [31:0] txdp;                          // Copy of descriptor pointer  
  reg [31:0] txdp_bu;                       // Backup value for TXDP
  reg [31:0] tx_desc_cache_link;            // Cached link of the Tx descriptor
  reg [31:0] tx_desc_cache_bufptr;          // Cached buffer pointer of the Tx descriptor
  reg [31:0] tx_desc_cache_cmdsts;          // Cached command/status fields
  reg [31:0] tx_desc_cache_extsts;          // Cached extended status fields 
  wire [15:0] tx_fifocnt_w;                 // Current amount of data in the tx_data_fifo in 32-bit words
  reg [15:0] tx_fifo_avail;                 // Current amount of free space in the tx_data_fifo in 32-bit words
                                            // (size of the tx_data_fifo - tx_fifo_cnt)
  reg ctdd;                                 // Flag, set when the current transmit descriptor has been completed
  reg [15:0] desc_cnt;                      // Count of bytes remaining in the current descriptor
  reg [15:0] desc_fifo_read;                // Number of dwords read from current descriptor into FIFO
  reg [15:0] desc_cnt_init;                 // Value of the desc_cnt at the beginning of the cycle 
  
  reg [31:0] frag_ptr;                      // Pointer to the next unread byte in the current fragment
  reg frame_done;                            // Flag, set when descriptor's data is read into FIFO 

  reg [15:0] word_count;                    // Word counter
  reg fifo_avail;
  reg frag_read;

  assign tx_desc_cnt_o = desc_cnt;
  wire [3:0] txfsm_cmd_w;
  wire [6:0] fsm_state_w;
  wire [7:0] flth = txcfg_i[15:8];          // Tx Fill Threshold in 32 bit words
  wire [7:0] drth = txcfg_i[7:0];           // Tx Drain Threshold in 32 bit words
  wire [2:0] mxdma = txcfg_i[22:20];        // Maximum size of transmit DMA data bursts                    
  assign tx_fsm_o = txfsm_cmd_w;
  
  assign tx_dread = txfsm_cmd_w[0] && ~txfsm_cmd_w[1] && ~txfsm_cmd_w[2] && ~txfsm_cmd_w[3]; 
  assign tx_fread = txfsm_cmd_w[0] && txfsm_cmd_w[1] && ~txfsm_cmd_w[2] && ~txfsm_cmd_w[3];
  assign tx_advance = ~txfsm_cmd_w[0] && ~txfsm_cmd_w[1] && txfsm_cmd_w[2] && ~txfsm_cmd_w[3];
  assign tx_dread_o = tx_dread || tx_fread;           // Any command which require data to be read from wb bus
  assign tx_dwrite_o = tx_advance;

  wire we_fifo_w =  ack_i && tx_fread && ~frame_done;    // If ~frame_done is not included one extra word is recorded to the FIFO
  reg we_fifo_pos;
  reg [15:0] timer_exp;
  reg timer_exp_r;
  
  assign test01_o = {fifo_avail,frame_done,frag_read,macxfer_done_i};

  // Modules instantiation
  tx_fsm tx_fsm_buf (
    .clk_i              (clk_i),
    .rst_i              (rst_i),
    .xfer_done_i        (xfer_done_i),
    .cr_txen_i          (cr_txen_i),
    .fifo_avail_i       (fifo_avail),
    .own_i              (tx_desc_cache_cmdsts[31]),
    .more_i             (tx_desc_cache_cmdsts[30]),
    .ctdd_i             (ctdd),    
    .frag_read_i        (frag_read),
    .frame_done_i       (frame_done),  
    .timer_exp_i        (timer_exp_r),
    .txfsm_cmd_o        (txfsm_cmd_w),
    .fsm_state_o        (fsm_state_w)
  );
  
  fifo tx_fifo (
    .wclk_i             (clk_i), 
    .rclk_i             (tx_clk_i), 
    .rst_i              (rst_i), 
    .we_i               (we_fifo_pos), 
    .rd_i               (tx_fread_i), 
    .data_i             (wb_data_i),
    .data_o             (tx_data_o), 
    .ffull_o            (tfifo_full_o), 
    .fempty_o           (tx_fempty_o),  
    .fifo_count_o       (tx_fifocnt_w),
    .wack_o             (),
    .rack_o             (),
    .afull_o            (),
    .aempty_o           () 
  );

 always @ (negedge clk_i)
   begin
     if (fsm_state_w[`TX_ADVANCE])
       begin
         if (~ack_i)
           wb_data_o <= tx_desc_cache_link;
         else
           begin
             case (word_count)                  // Synthesis full case
               16'h0: wb_data_o <= tx_desc_cache_bufptr;
               16'h1: wb_data_o <= (tx_desc_cache_cmdsts & 32'h7FFFFFFF) | 32'h08000000;    // Clear OWN bit and set OK bit - packet Tx'd successfully
               16'h2: wb_data_o <= tx_desc_cache_extsts;
             endcase
           end
       end
   end
  
  always @ (tx_dread or tx_fread or tx_advance)
    begin
      if (tx_dread)
        begin
          wb_adr_o = txdp;
          data_count_o = `RD_DESC;
        end
      else if (tx_fread)
        begin
          wb_adr_o = tx_desc_cache_bufptr;
          data_count_o = desc_cnt_init;
        end
      else if (tx_advance)    // Descriptor needs to be written back to the main memory
        begin
          wb_adr_o = txdp_bu;
          data_count_o = `RD_DESC;
        end
    end
    
    
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          tx_start_o <= 1'b0;
        end
      else
        begin
          if (tx_fifo_avail > {8'h0,flth})
            fifo_avail <= 1'b1;
          else
            fifo_avail <= 1'b0;
            
          if ({8'h00,drth} < (`TX_FIFO_SIZE - tx_fifo_avail))   // Start transmission when either FIFO threshold is reached
            tx_start_o <= 1'b1;                                 // or full packet has been received
          else if (xfer_done_i && fsm_state_w[`TX_FRAG_READ])
            tx_start_o <= 1'b1;
          else
            tx_start_o <= 1'b0;
        end
    end 
    
  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          ctdd <= 1'b0;
          frame_done <= 1'b0; 
          timer_exp <= 16'h0;
          timer_exp_r <= 1'b0;
        end
      else
        begin
          if (txdpw_i)
            begin
              ctdd <= 1'b0;       // New TXDP has been loaded by the driver
              txdp <= txdp_i;   // TXDP updated by the driver
            end
          else if ((word_count == 16'h2) && fsm_state_w[`TX_ADVANCE] && (tx_desc_cache_link == 32'h0))
            ctdd <= 1'b1;       // Current TXDP has been completed
        end
      case (1'b1) 
        fsm_state_w[`TX_IDLE]:
          begin					
            if (timer_exp == `TX_TIMER_EXP)
              timer_exp_r <= 1'b1;
            else
              begin
                timer_exp <= timer_exp + 1'b1;
                timer_exp_r <= 1'b0;
              end
            frame_done <= 1'b0;
            word_count <= 16'h0;
          end        
        fsm_state_w[`TX_DESC_READ]:  // One Tx descriptor needs to be read from memory
          begin
            txdp_bu <= txdp; 
            timer_exp <= 16'h0;
            timer_exp_r <= 1'b0;
            case ({word_count,ack_i})   // Synthesis full case
              {16'h0,1'b1}: tx_desc_cache_link <= wb_data_i;
              {16'h1,1'b1}: tx_desc_cache_bufptr <= wb_data_i;
              {16'h2,1'b1}: tx_desc_cache_cmdsts <= wb_data_i;
              {16'h3,1'b1}: tx_desc_cache_extsts <= wb_data_i;
            endcase
            if (xfer_done_i)
              begin
                frag_read <= 1'b0;
                desc_cnt <= tx_desc_cache_cmdsts[15:0]; 
                desc_cnt_init <= tx_desc_cache_cmdsts[15:0];
                desc_fifo_read <= {2'b00,tx_desc_cache_cmdsts[15:2]};   // Number of dwords
              end
            if (ack_i)
              word_count <= word_count + 1'b1;
          end
        fsm_state_w[`TX_FIFO_BLOCK]:
          begin
            if (macxfer_done_i == 1'b1)
              begin
                desc_cnt <= 16'h0;
                frag_read <= 1'b1;
              end
            word_count <= 16'h0;
          end
        fsm_state_w[`TX_FRAG_READ]:
          begin
            if (we_fifo_pos)
              desc_fifo_read <= desc_fifo_read - 1'b1;
            if (desc_fifo_read == 16'h1)
              frame_done <= 1'b1;
          end 
        fsm_state_w[`TX_ADVANCE]:
          begin
            timer_exp <= `TX_TIMER_EXP;
            if (ack_i)
              word_count <= word_count + 1'b1;
            if (tx_desc_cache_link != 32'h0)
              txdp <= tx_desc_cache_link;              
          end 
      endcase
      
    end
    
    always @ (negedge clk_i)
      begin
        if (!rst_i)
          begin
            tx_fifo_avail <= `TX_FIFO_SIZE;
          end
        else
          begin
            tx_fifo_avail <= `TX_FIFO_SIZE - tx_fifocnt_w;
          end
      end
      
  always @ (posedge clk_i)
    begin
      we_fifo_pos <= we_fifo_w;
    end
  
endmodule
