//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// TX state machine
// 
`include "parms.v"

module tx_fsm (
  clk_i, rst_i, xfer_done_i, cr_txen_i, fifo_avail_i, own_i, more_i,
  ctdd_i, frag_read_i, frame_done_i, timer_exp_i,
  txfsm_cmd_o, fsm_state_o
  
  );
  
  input clk_i;                    // Input clock signal
  input rst_i;                    // Input reset signal
  input xfer_done_i;              // Completion of a PCI bus transfer request
  input cr_txen_i;                // TX enable bit from CR operational register, changes on the negedge
  input fifo_avail_i;             // TxFIFOAvail is greater than TxFillThreshold, changes of the negedge
  input own_i;                    // Descriptor OWN bit
  input more_i;                   // MORE bit from the cmdsts field of the current descriptor
  input ctdd_i;                   // Flag, set when the current transmit descriptor has been completed
  input frag_read_i;              // Signal from Tx bufmgr that fragment is read - desc counter is 0
  input frame_done_i;             // Signal that all the data from current descriptor has been read into FIFO
  input timer_exp_i;              // Timer to wait before going to the TX_DESC_READ state
  output [3:0] txfsm_cmd_o;       // Output FSM command
  output [6:0] fsm_state_o;       // Output FSM state
    
    // FSM registers and wires
  reg [6:0] tx_fsm_state;                   // Current state of the FSM
  reg [6:0] tx_fsm_next_state;              // Next state of the FSM
  reg [3:0] tx_fsm_out;                     // Output command register of the FSM
  reg xfer_done_buf;
  

  assign fsm_state_o = tx_fsm_state;
  assign txfsm_cmd_o = tx_fsm_out;
                                        
  // Signals in the sensitivity list change on the negedge            
  always @ (tx_fsm_state or ctdd_i or cr_txen_i or xfer_done_buf or own_i or fifo_avail_i
            or more_i or frag_read_i or frame_done_i or timer_exp_i)   // Block for the FSM states
    begin
      tx_fsm_next_state = 7'b0;
      tx_fsm_out = `TX_NOP;         // All commands asserted on the posedge
      case (1'b1)           
        tx_fsm_state[`TX_IDLE]:
          begin
            tx_fsm_out = `TX_NOP;
            if (cr_txen_i && !ctdd_i && timer_exp_i)
              tx_fsm_next_state[`TX_DESC_READ] = 1'b1;
            else
              tx_fsm_next_state[`TX_IDLE] = 1'b1;       // Otherwise stay in idle state
          end
        tx_fsm_state[`TX_DESC_READ]:
          begin
            tx_fsm_out = `TX_DREAD;
            if (xfer_done_buf && own_i)
              tx_fsm_next_state[`TX_FIFO_BLOCK] = 1'b1;
            else if (xfer_done_buf && !own_i)
              tx_fsm_next_state[`TX_IDLE] = 1'b1;
            else
              tx_fsm_next_state[`TX_DESC_READ] = 1'b1;        // Otherwise stay in the same position
          end
        tx_fsm_state[`TX_FIFO_BLOCK]:     // Waiting for free space in the TxFIFO to reach TxFillThreshold
          begin
            tx_fsm_out = `TX_FBLOCK;
            if (fifo_avail_i && ~frame_done_i && ~frag_read_i)
              begin
                tx_fsm_next_state[`TX_FRAG_READ] = 1'b1;
              end
            else if (frag_read_i && !more_i)
              begin                                                           
                tx_fsm_next_state[`TX_ADVANCE] = 1'b1;
              end
            else 
                tx_fsm_next_state[`TX_FIFO_BLOCK] = 1'b1;         // Otherwise stay in the same state
          end
        tx_fsm_state[`TX_FRAG_READ]:      // FIFO is available - need to read block of data
          begin
            tx_fsm_out = `TX_FREAD;
            if (xfer_done_buf)
              begin
                tx_fsm_next_state[`TX_FIFO_BLOCK] = 1'b1;
              end
            else
              begin
                tx_fsm_next_state[`TX_FRAG_READ] = 1'b1;
              end
          end
        tx_fsm_state[`TX_ADVANCE]:        // Need to update current register and load new if link field is not NULL
          begin
            tx_fsm_out = `TX_DADVANCE;
            if (xfer_done_buf)
              tx_fsm_next_state[`TX_IDLE] = 1'b1;
            else
              tx_fsm_next_state[`TX_ADVANCE] = 1'b1;
          end
      endcase
    end
 
 always @ (posedge clk_i)
   begin
     if (!rst_i)            // Reset condition
       begin
         tx_fsm_state <= 7'b0;
         tx_fsm_state[`TX_IDLE] <= 1'b1;
       end
     else
       begin
         tx_fsm_state <= tx_fsm_next_state;
       end
   end
   
  
  always @ (negedge clk_i)
    begin
      xfer_done_buf <= xfer_done_i;
    end
    
endmodule
 
 