//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Top module to communicate with PHY
//
`include "parms.v"

module tx_mac (
  clk_i, rst_i, data_i, tx_clk_i, phy_col_i,
  phy_crs_i, link_speed_i, duplex_i, tx_start_i,
  tx_desc_cnt_i, tx_fempty_i, gtx_clk_i,
  data_o, tx_en_o, tx_er_o, tx_fread_o,
  tx_clk_o, xfer_done_o, xfer_failed_o, txmac_fsm_o, byte_cnt_o  
  );
  
  input clk_i;                    // Input clock signal
  input rst_i;                    // Input reset signal
  input [31:0] data_i;            // Input data from the Tx buffer manager
  input tx_clk_i;                 // Input Tx clock for 10/100Mb modes from PHY
  input phy_col_i;                // Collision detection from PHY
  input phy_crs_i;                // Carrier sense signal from PHY - asserted when media is not idle
  input [2:0] link_speed_i;       // Link speed input from PHY
  input duplex_i;                 // Duplex mode input from PHY
  input tx_start_i;               // Transmit start signal from tx_bufmgr
  input [15:0] tx_desc_cnt_i;     // Total number of bytes in the current descriptor to be Tx'd
  input tx_fempty_i;              // Signal from tx_bufmgr that FIFO is empty
  input gtx_clk_i;                // Input clock signal for 1G mode
  output [7:0] data_o;            // Output data to the PHY
  output tx_en_o;                 // Tx enable signal to PHY
  output tx_er_o;                 // Tx error signal to PHY
  output tx_fread_o;              // Signal to tx_bufmgr to read 32-bit word from it's FIFO - asserted on the posedge
  output tx_clk_o;                // Output tx clock to syncronize Tx FIFO
  output xfer_done_o;             // Signal to txbuf_mgr that transmission has been completed successfully, asserted on the posedge
  output xfer_failed_o;           // Signal to txbuf_mgr that transmission failed, asserted on the posedge
  output [3:0] txmac_fsm_o;             // Output for monitoring purposes
  output [15:0] byte_cnt_o;       // Output for monitoring purposes
   
  reg tx_en_o;
  reg tx_er_o;
  reg [7:0] data_o;
  reg tx_fread_o;
  
  // Internal wires and registers
  reg [15:0] byte_cnt;      // Bytes counter
  reg [15:0] tx_desc;
  reg [1:0] word_load;      // Register to count when 32-bit word to be loaded from FIFO 
  reg [2:0] crc_count;      // Register to count crc bytes
  reg [15:0] iframe_gap;
  reg xfer_done;
  reg xfer_failed;
  reg xfer_done_o;
  reg xfer_failed_o;	
  reg next_state1;
  reg next_state2;	 
  reg tx_pad;

  
  wire tx_clk = (link_speed_i == 3'b100) ? gtx_clk_i : tx_clk_i;     // MUX to select Tx clock
  assign tx_clk_o = gtx_clk_i;
  
  reg crc_init_r;
  reg crc_dataen_r;
  reg crc_rd_r;
  wire [7:0] crco_w;											   
  assign byte_cnt_o = {tx_start_i,next_state1,next_state2,duplex_i,phy_crs_i,tx_pad,2'b00,byte_cnt[7:0]};
  
  // Modulese instantiation
  crc_gen_tx crc_gen_txmac (
    .clk_i                (gtx_clk_i), 
    .rst_i                (rst_i),
    .init_i               (crc_init_r), 
    .frame_data_i         (data_o), 
    .data_en_i            (crc_dataen_r), 
    .crc_rd_i             (crc_rd_r),
    .crc_o                (crco_w)
  );
  
//-------------------------------Tx MAC FSM definition--------------------------
  reg [3:0] txmac_fsm_state;
  reg [3:0] txmac_fsm_next_state;								   
  assign txmac_fsm_o = txmac_fsm_state;
  
  always @ (txmac_fsm_state or tx_start_i or next_state1 or next_state2   // All signals in sensitivity list (except from txmac_fsm_state) must be changing on the posedge
            or duplex_i or phy_crs_i or tx_pad)
    begin
      txmac_fsm_next_state = `TXMAC_IDLE;
      xfer_done = 1'b0;
      xfer_failed = 1'b0;
      case (txmac_fsm_state)            // Synthesis full case
        `TXMAC_IDLE:
          begin
            if (tx_start_i)
              txmac_fsm_next_state = `TXMAC_START;
            else
              txmac_fsm_next_state = `TXMAC_IDLE;
          end
        `TXMAC_START:
          begin
            if (duplex_i && ~phy_crs_i)       // No collision detection is required
              txmac_fsm_next_state = `TXMAC_PREAMB;
            else
              txmac_fsm_next_state = `TXMAC_START;        
          end
        `TXMAC_PREAMB:    //  7 bytes of preamble and one byte SFD to be transmitted
          begin
            if (next_state1)
              txmac_fsm_next_state = `TXMAC_FDSTART;
            else
              txmac_fsm_next_state = `TXMAC_PREAMB;       // Otherwise stay in the same state        
          end
        `TXMAC_FDSTART:
          begin
            if (next_state2)
              begin
                if (tx_pad)
                  txmac_fsm_next_state = `TXMAC_PAD;    // If data size is less than 60 bytes padding needs to be inserted
                else
                  txmac_fsm_next_state = `TXMAC_FCS;
              end
            else        
              txmac_fsm_next_state = `TXMAC_FDSTART;  
          end
        `TXMAC_FCS:
          begin
            if (next_state1 & next_state2)
              begin
                txmac_fsm_next_state = `TXMAC_IFRAME;
              end
            else
              txmac_fsm_next_state = `TXMAC_FCS;     
          end
        `TXMAC_PAD:
          begin
            if (next_state1)      // Wait until padding is filled
              txmac_fsm_next_state = `TXMAC_FCS;
            else
              txmac_fsm_next_state = `TXMAC_PAD;
          end
        `TXMAC_IFRAME:
          begin
            xfer_done = 1'b1;
            if (~next_state1)
              txmac_fsm_next_state = `TXMAC_IDLE;
            else        
              txmac_fsm_next_state = `TXMAC_IFRAME;        
          end
      endcase
    end

  always @ (negedge gtx_clk_i)
    begin
      if (!rst_i)
        begin
          txmac_fsm_state <= `TXMAC_IDLE;
        end
      else
        begin
          txmac_fsm_state <= txmac_fsm_next_state;
        end
    end


//----------------------------End of Tx MAC FSM definition----------------------
  
 always @ (posedge gtx_clk_i)
   begin
     xfer_done_o <= xfer_done;
     xfer_failed_o <= xfer_failed;
   end                                                                                        
   
 always @ (negedge gtx_clk_i)            // Was timing problem in Plan ahead with posedge
   begin
     if (!rst_i)
       begin
         tx_fread_o <= 1'b0;
         crc_dataen_r <= 1'b0;
         crc_rd_r <= 1'b0;	
         next_state1 <= 1'b0;
         next_state2 <= 1'b0;
         tx_en_o <= 1'b0;
         tx_er_o <= 1'b0;
       end
     else
       begin  
         if (txmac_fsm_state == `TXMAC_IDLE)
           begin
             byte_cnt <= 16'h0;  
             tx_en_o <= 1'b0;
             tx_er_o <= 1'b0; 
             crc_init_r <= 1'b1;
           end					
         else if (txmac_fsm_state == `TXMAC_START)
           begin
             next_state1 <= 1'b0; 
           end
         else if (txmac_fsm_state == `TXMAC_PREAMB)
           begin		 
             tx_en_o <= 1'b1;		
             if (byte_cnt == 16'h7)
               byte_cnt <= 16'h1;
             else
               byte_cnt <= byte_cnt + 1'b1;
             case (byte_cnt)            // Synthesis full case
               16'h6:
                 begin
                   next_state1 <= 1'b1;
                   next_state2 <= 1'b0;
                   data_o <= `PREAMBLE_PATTERN;
                   tx_fread_o <= 1'b1;                 
                 end 
               16'h7:
                 begin
                   data_o <= `SFD_PATTERN;
                   word_load <= 2'b00;
                   tx_desc <= tx_desc_cnt_i - 1'b1;
                   tx_fread_o <= 1'b0;                 
                 end
               default:
                 begin
                   data_o <= `PREAMBLE_PATTERN;                  
                 end
             endcase
           end         
         else if (txmac_fsm_state == `TXMAC_FDSTART)   // Start main data transmission
           begin
             crc_init_r <= 1'b0;
             crc_dataen_r <= 1'b1;
             word_load <= word_load + 1'b1;
             byte_cnt <= byte_cnt + 1'b1;
             next_state1 <= 1'b0;
             if (tx_desc < `MIN_PACKET_SIZE)
               tx_pad <= 1'b1;
             else
               tx_pad <= 1'b0;
             if ((word_load == 2'b10) && (byte_cnt != (tx_desc + 1'b1)))  // Second comparison is required to cut
               tx_fread_o <= 1'b1;                             // last (extra) FIFO read
             else 
               tx_fread_o <= 1'b0;
             if (word_load == 2'b00)
               data_o <= data_i[31:24];
             else if (word_load == 2'b01)
               data_o <= data_i[23:16];
             else if (word_load == 2'b10)
               data_o <= data_i[15:8];
             else if (word_load == 2'b11)
               data_o <= data_i[7:0];
             
             if (byte_cnt == tx_desc)
               next_state2 <= 1'b1;
             crc_count <= 3'b000;
           end
         else if (txmac_fsm_state == `TXMAC_FCS)      // CRC calculation
           begin
             crc_count <= crc_count + 1'b1;
             data_o <= crco_w;
             crc_dataen_r <= 1'b0;
             if (crc_count == 3'b100) 
               begin
                 crc_rd_r <= 1'b0;
                 tx_en_o <= 1'b0;
               end  
             else
               crc_rd_r <= 1'b1;

             if (crc_count == 3'b011)
               begin
                 next_state1 <= 1'b1;
                 next_state2 <= 1'b1;
               end              
               
              if (link_speed_i == 3'b100)       // Gigabit ethernet
                iframe_gap <= `IFRAME_GAP1000;
              else if (link_speed_i == 3'b010)   //100M ethernet
                iframe_gap <= `IFRAME_GAP100;
              else                              // Just in case
                iframe_gap <= 16'h20;
                  
           end
         else if (txmac_fsm_state == `TXMAC_PAD)      // Padding needs to be inserted
           begin
             byte_cnt <= byte_cnt + 1'b1;
             data_o <= `PAD_PATTERN;
             crc_count <= 2'b00; 
             if (byte_cnt == `MIN_PACKET_SIZE)
               begin
                 next_state1 <= 1'b1;
                 next_state2 <= 1'b0;
               end
           end
         else if (txmac_fsm_state == `TXMAC_IFRAME)
           begin
             iframe_gap <= iframe_gap - 1'b1;
             if (iframe_gap == 16'h0)
               begin
                 next_state1 <= 1'b0;
               end
           end
       end    
   end

endmodule
