//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Wishbone controller module for the gigabit ethernet controller

`include "parms.v"

module wb_cntr_g (
  clk_i, rst_i, rd_i, int_wr_i, wcntr_i, dataex_i, adrex_i, ack_i, int_adr_i, int_data_i,
  stb_i, cyc_i, sel_i, we_i,
  dataex_o, adrex_o, we_o, sel_o, stb_o, cyc_o, int_data_o, xfer_done_o,
  data_ready_o, int_adr_o, we_opreg_o, ack_o, fsm_state_o
  );
  
  input clk_i;              // Input clock
  input rst_i;              // Input reset signal
  input rd_i;               // Request to read data from the memory, asserted on the posedge
  input int_wr_i;           // Request from internal devices to write data to external memory
  input [15:0] wcntr_i;     // Number of bytes to be read, asserted on the posedge
  input [31:0] dataex_i;    // External data input
  input [31:0] adrex_i;     // External address input
  input ack_i;              // Input acknowledge signal
  input [31:0] int_adr_i;   // Address input from internal logic
  input [31:0] int_data_i;  // Internal data input
  input stb_i;              // Input strobe
  input cyc_i;              // Input cycle signal
  input [3:0] sel_i;        // Input byte select signal
  input we_i;               // Input write enable signal
  output [31:0] dataex_o;   // External data output
  output [31:0] adrex_o;    // External address output
  output we_o;              // Output write enable signal
  output [3:0] sel_o;       // Byte enable output
  output stb_o;             // Output strobe
  output cyc_o;             // Bus cycle
  output [31:0] int_data_o; // Output data for internal consumers
  output xfer_done_o;       // Signal that data transfer is completed
  output data_ready_o;      // Signal for internal consumers that data is ready
  output [31:0] int_adr_o;  // Output address for internal devices
  output we_opreg_o;        // Write enable for operations registers - changes on the posedge
  output ack_o;             // Output acknowledge signal - changes on the negedge
  output [3:0] fsm_state_o; // State of FSM
  
  reg [31:0] adrex_o;
  reg we_o;
  reg [3:0] sel_o;
  reg stb_o;
  reg cyc_o;
  reg [31:0] int_data_o;
  reg xfer_done_o;
  reg we_opreg_o;
  reg ack_o;
  
  reg [3:0] wb_fsm_state;         // Current state of wb FSM
  reg [3:0] wb_fsm_next_state;    // Next state of wb FSM
  reg [3:0] wb_fsm_out;           // Output of wb FSM
  
  reg [15:0] wcount;              // Register to store number of 32-bit words to be read
  reg [15:0] bus_counter;           // Bus counter
  reg next_state1;  
    
  assign fsm_state_o = wb_fsm_state;
  assign data_ready_o = ack_i;
  assign dataex_o = int_data_i;
  
  //------------------------------------WB FSM------------------------------------------
  always @ (wb_fsm_state or rd_i  or sel_i or we_i or cyc_i or stb_i
             or int_wr_i or xfer_done_o)            // Latches definition for the wb FSM
    begin
      wb_fsm_next_state = `WB_IDLE;
      wb_fsm_out = `WB_IDLE;
      ack_o = 1'b0;
      we_opreg_o = 1'b0;
      case (wb_fsm_state)
        `WB_IDLE:     // IDLE state
          begin
            if (int_wr_i)
              wb_fsm_next_state = `WB_BWRITE;
            else if (rd_i)
              wb_fsm_next_state = `WB_BREAD; 
           else if (we_i && cyc_i && stb_i)     // External write from wb_host
             begin
               wb_fsm_next_state = `WB_EXWRITE;
             end
           else
             wb_fsm_next_state = `WB_IDLE;  
          end
         `WB_BREAD:   // Data burst to be read
           begin
             if (xfer_done_o)           // Read of the burst is completed
               wb_fsm_next_state = `WB_IDLE;
             else
               wb_fsm_next_state = `WB_BREAD;   // Otherwise stay in the same state
           end
         `WB_EXWRITE:                 // External write form wb_host
           begin
             if (~cyc_i)      // Wait until cycle is completed
               begin
                 wb_fsm_next_state = `WB_IDLE;
               end
             else
               begin
                 wb_fsm_next_state = `WB_IDLE;
               end
             we_opreg_o = 1'b1;
             ack_o = 1'b1;
           end
          `WB_BWRITE:         // Write data to the external memory
            begin
              if (xfer_done_o)
                wb_fsm_next_state = `WB_IDLE;
              else
                wb_fsm_next_state = `WB_BWRITE;
            end
      endcase
    end
  
  always @ (negedge clk_i)          // Registers definition for the wb FSM
    begin
      if (!rst_i)
        begin
          wb_fsm_state <= `WB_IDLE;
        end
      else
        begin
          wb_fsm_state <= wb_fsm_next_state;
        end
    end
  
//-------------------------------------End of WB FSM-------------------------------------
// 
// Buffers for FSM outputs

  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          we_o <= 1'b0;         
          sel_o <= 4'h0;
          stb_o <= 1'b0;
          cyc_o <= 1'b0;
          bus_counter <= 16'h0;          
        end
      else
        begin
          case (wb_fsm_state)
            `WB_IDLE: 
              begin 
                xfer_done_o <= 1'b0;
                bus_counter <= 16'h0;
                we_o <= 1'b0;  
                wcount <= (wcntr_i >> 2);     // Otherwise FSM flips to IDLE until wcount is received      
              end
            `WB_BREAD:
              begin
                bus_counter <= bus_counter + 1'b1;
                if (bus_counter == 16'h0)         // Load word count at the first cycle of burst read
                  begin
                    wcount <= (wcntr_i >> 2);
                    adrex_o <= int_adr_i;
                    we_o <= 1'b0;
                    sel_o <= 4'hF;          // Read 32-bit word
                    stb_o <= 1'b1;
                    cyc_o <= 1'b1;
                  end
                else
                  begin
                    if (wcount == 16'h0)     // Read cycle is complete
                      begin
                        stb_o <= 1'b0;
                        cyc_o <= 1'b0;
                        xfer_done_o <= 1'b1;
                      end
                    else if (ack_i)               // Read cycle is ongoing
                      begin
                        wcount <= wcount - 1'b1;    // If acknowledge is received read one word
                        adrex_o <= adrex_o + 4'h4;
                        int_data_o <= dataex_i;
                      end
                  end    
              end
            `WB_BWRITE:
              begin
                bus_counter <= bus_counter + 1'b1;
                if (bus_counter == 16'h0)
                  begin
                    wcount <= (wcntr_i >> 2);
                    adrex_o <= int_adr_i;
                  end
                else
                  begin
                    we_o <= 1'b1;
                    sel_o <= 4'hF;
                    stb_o <= 1'b1;
                    cyc_o <= 1'b1;
                    if (wcount == 16'h1)
                      begin
                        stb_o <= 1'b0;
                        cyc_o <= 1'b0;
                        we_o <= 1'b0;
                        xfer_done_o <= 1'b1;
                      end
                      
                    if (ack_i)
                      begin
                        wcount <= wcount - 1'b1;
                        adrex_o <= adrex_o + 4'h4;
                      end
                  end                  
              end
          endcase
        end  
    end
      
endmodule
